@@ -1,52 +1,58 @@
 package App::ArduinoBuilder;
 
-use 5.022;
+use 5.026;
 use strict;
 use warnings;
 use utf8;
 
 use App::ArduinoBuilder::Builder 'build_archive', 'build_object_files', 'link_executable', 'run_hook';
+use App::ArduinoBuilder::CommandRunner;
 use App::ArduinoBuilder::Config 'get_os_name';
 use App::ArduinoBuilder::FilePath 'find_latest_revision_dir', 'list_sub_directories', 'find_all_files_with_extensions';
 use App::ArduinoBuilder::Logger;
 use App::ArduinoBuilder::System 'find_arduino_dir', 'system_cwd';
 
 use File::Basename;
+use File::Path 'remove_tree';
 use File::Spec::Functions;
 use Getopt::Long;
 use List::Util 'any', 'none', 'first';
 use Pod::Usage;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 sub Run {
   my $project_dir;
   my $build_dir;
 
+  my $config = App::ArduinoBuilder::Config->new();
+
   my (@skip, @force, @only);
   GetOptions(
       'help|h' => sub { pod2usage(-exitval => 0, -verbose => 2)},
       'project-dir|project|p=s' => \$project_dir,
       'build-dir|build|b=s' => \$build_dir,
       'log-level|l=s' => sub { App::ArduinoBuilder::Logger::set_log_level($_[1]) },
+      'config|c=s%' => sub { $config->set($_[1] => $_[2], allow_override => 1) },
+      'menu=s%' => sub { $config->set('builder.menu.'.$_[1] => $_[2], allow_override => 1) },
       'skip=s@' => sub { push @skip, split /,/, $_[1] },  # skip this step
       'force=s@' => sub { push @force, split /,/, $_[1] },  # even if it would be skipped by the dependency checker
       'only=s@' => sub { push @only, split /,/, $_[1] },  # run only these steps (skip all others)
       'stack-trace-on-error|stack' => sub { App::ArduinoBuilder::Logger::print_stack_on_fatal_error(1) },
+      'j=i' => sub { $config->set('builder.parallelize' => $_[1], allow_override => 1) },
     ) or pod2usage(-exitval => 2, -verbose =>0);
 
+  fatal "More than one command specified: ".join(' ', @ARGV) if @ARGV > 1;
+  my $command = @ARGV ? $ARGV[0] : 'build';
+
   my $project_dir_is_cwd = 0;
   if (!$project_dir) {
     $project_dir_is_cwd = 1;
     $project_dir = system_cwd();
   }
 
-  my $config = App::ArduinoBuilder::Config->new(
-      files => [catfile($project_dir, 'arduino_builder.local'),
-                catfile($project_dir, 'arduino_builder.config')],
-      allow_missing => 1,
-      allow_partial => 1,
-      resolve => 1);
+  $config->read_file(catfile($project_dir, 'arduino_builder.local'), allow_missing => 1);
+  $config->read_file(catfile($project_dir, 'arduino_builder.config'), allow_missing => 1);
 
   $config->set('builder.project_dir' => $project_dir);
 
@@ -60,6 +66,48 @@ sub Run {
     }
   }
 
+  if ($command eq 'build') {
+    build($config, $build_dir, \@skip, \@force, \@only);
+  } elsif ($command eq 'clean') {
+    clean($config, $build_dir);
+  } else {
+    fatal "Unknown command: ${command}";
+  }
+}
+
+sub clean {
+  my ($config, $build_dir) = @_;
+
+  # TODO: add a way to clean only parts of the projects.
+  if ($build_dir eq $config->get('builder.default_build_dir')) {
+    remove_tree($build_dir, {safe => 1, keep_root => 1});
+  } else {
+    warning "For safety reason we can only clean build directory specified in the project";
+    warning "config. You should run `rm -rf` manually.";
+    fatal "Not cleaning context dependent directory: ${build_dir}";
+  }
+}
+
+sub build {
+  my ($config, $build_dir, @array_args) = @_;
+  my @skip = @{$array_args[0]};
+  my @force = @{$array_args[1]};
+  my @only = @{$array_args[2]};
+
+  if (!$config->exists('builder.source.path')) {
+    my $project_dir = $config->get('builder.project_dir');
+    my $d = first { -d catdir($project_dir, $_) } qw(src srcs source sources);
+    if (defined $d) {
+      $config->set('builder.source.path' => catdir($project_dir, $d));
+      $config->set('builder.source.is_recursive' => 1, ignore_existing => 1);
+    } else {
+      $config->set('builder.source.path' => $project_dir);
+      $config->set('builder.source.is_recursive' => 0, ignore_existing => 1);
+    }
+  } else {
+    $config->set('builder.source.is_recursive' => 1, ignore_existing => 1);
+  }
+
   if (!$config->exists('builder.package.path')) {
     fatal 'At least one of builder.package.path or builder.package.name must be specified in the config' unless $config->exists('builder.package.name');
     # TODO: the core package can also be installed in a "hardware" directory in
@@ -94,33 +142,40 @@ sub Run {
       fatal 'The builder.package.arch config is not set and more than one arch is present in the package: '.$hardware_dir;
     }
   }
-  my $hardware_path = find_latest_revision_dir(catdir($hardware_dir, $config->get('builder.package.arch')));
-
   debug "Project config: \n%s", sub { $config->dump('  ') };
 
-  my $boards_local_config_path = catfile($hardware_path, 'boards.local.txt');
-  if (-f $boards_local_config_path) {
-    my $board_name = $config->get('builder.package.board');
-    my $board = App::ArduinoBuilder::Config->new(file => $boards_local_config_path, resolve => 1, allow_partial => 1)->filter($board_name);
-    $config->merge($board);
-  }
+  my $hardware_path = find_latest_revision_dir(catdir($hardware_dir, $config->get('builder.package.arch')));
 
-  my $boards_config_path = catfile($hardware_path, 'boards.txt');
-  if (-f $boards_config_path) {
-    my $board_name = $config->get('builder.package.board');
-    my $board = App::ArduinoBuilder::Config->new(file => $boards_config_path, resolve => 1, allow_partial => 1)->filter($board_name);
-    fatal "Board '${board_name}' not found in boards.txt." if $board->empty();
-    $config->merge($board);
-  } else {
-    warning "Could not find boards.txt file.";
-  }
+  my $all_boards_config = App::ArduinoBuilder::Config->new(
+      files => [catfile($hardware_path, 'boards.local.txt'),
+                catfile($hardware_path, 'boards.txt')],
+      allow_missing => 1);
+  warning "Could not find any board config file" unless $all_boards_config->nb_files();
+  my $board_name = $config->get('builder.package.board');
+  my $board_config = $all_boards_config->filter($board_name);
+  # In $all_board_config we may have interesting values (or maybe even important
+  # ones – altough not with the core I checked). In particular the name of each
+  # menu. However, merging it here result in a dump that is far too large when
+  # we are in full debug mode.
+  #$board_config->merge($all_boards_config);
 
+  # We should check whether a menu can be configured in the platform.txt, if so
+  # this block should be moved to below (and platform.txt should be merged into
+  # what is currently called board_config).
   my $menu_config = $config->filter('builder.menu');
-  my $board_menu = $config->filter('menu');
+  my $board_menu = $board_config->filter('menu');
   for my $m ($menu_config->keys()) {
     my $v = $menu_config->get($m);
-    $config->merge($board_menu->filter("${m}.${v}"));
+    # This is one of the rare case where we want a merge to override previously
+    # existing config. Although we still don’t want to override config set
+    # directly by the user (e.g. on the command line), which is way all this is
+    # done in a temporary config.
+    my $menu_value = $board_menu->filter("${m}.${v}");
+    full_debug "Merging menu values for menu '${m}' with key '${v}':\n%s", sub { $menu_value->dump('  ') };
+    $board_config->merge($menu_value, allow_override => 1);
   }
+  $config->merge($board_config);
+  # TODO: warn about unset menus
 
   # TODO: Handles core, variant and tools references:
   # https://arduino.github.io/arduino-cli/0.32/platform-specification/#core-reference
@@ -132,18 +187,15 @@ sub Run {
   $config->set('runtime.platform.path' => $hardware_path);
   # Unclear what the runtime.hardware.path variable is supposed to point to.
   $config->set('runtime.os' => get_os_name());
-  $config->set('runtime.ide.version' => '2040');  # The version that we are emulating currently (2.4.0).
-  $config->set('ide_version' => '2040');
+  $config->set('runtime.ide.version' => '10607');
+  $config->set('ide_version' => '10607');
   $config->set('software' => 'ARDUINO');
   # todo: name, _id, build.fqbn, and the time options
-  # Note: build.source.path is the root of the source. So, if we build from an
-  # src sub-directory (for example), it should be that directory and not the
-  # root of the project directory.
-  $config->set('build.source.path' => $project_dir);
-  $config->set('sketch_path' => $project_dir);
+  $config->set('build.source.path' => $config->get('builder.source.path'));
+  $config->set('sketch_path' => $config->get('builder.source.path'));
   $config->set('build.path' => $build_dir);
   $config->set('build.project_name' => $config->get('builder.project_name'));
-  $config->set('build.arch' => $config->get('builder.package.arch'));
+  $config->set('build.arch' => uc($config->get('builder.package.arch')));  # Undocumented but it seems that it’s always upper case.
   $config->set('build.core.path', catdir($hardware_path, 'cores', $config->get('build.core')));
   $config->set('build.system.path', catdir($hardware_path, 'system'));
   $config->set('build.variant.path', catdir($hardware_path, 'variants', $config->get('build.variant')));
@@ -163,16 +215,21 @@ sub Run {
     }
   }
 
+  my $config_append = $config->filter('builder.config.append');
+  for my $k ($config_append->keys()) {
+    $config->append($k, $config_append->get($k));
+  }
+
   # TODO: we should create config for all the tools defined by all the other
   # platform (not overriding the existing definitions). There are some other
   # considerations that we are not handling yet from:
   # https://arduino.github.io/arduino-cli/0.32/package_index_json-specification/#how-a-tools-path-is-determined-in-platformtxt
 
-  # TODO: we should probably never call resolve because variables can change
-  # but the config could have a cache of resolved values with invalidation on
-  # the right variable change. BUT we still need to do the os_name suffix
-  # resolution.
-  $config->resolve(allow_partial => 1);
+  full_debug "Complete configuration: \n%s", sub { $config->dump('  ') };
+
+  if ($config->exists('builder.parallelize')) {
+    default_runner()->set_max_parallel_tasks($config->get('builder.parallelize'));
+  }
 
   my $run_step = sub {
     my ($step) = @_;
@@ -254,12 +311,16 @@ sub Run {
     }
   }
 
-  $config->append('includes', '"-I'.$config->get('build.source.path').'"');
+  $config->append('includes', '"-I'.$config->get('builder.source.path').'"');
 
   if ($run_step->('sketch')) {
     info 'Building sketch...';
     $builder->run_hook('sketch.prebuild');
-    my $built_sketch = $builder->build_object_files($config->get('build.source.path'), catdir($build_dir, 'sketch'), [$build_dir], $force->('sketch'));
+    # TODO: add configuration option for the ignored directories and also a way to
+    # build only the code inside the src/ directory
+    my $built_sketch = $builder->build_object_files(
+        $config->get('builder.source.path'), catdir($build_dir, 'sketch'),
+        [], $force->('sketch'), $config->get('builder.source.is_recursive'));
     info ($built_sketch ? '  Success' : '  Already up-to-date');
     $built_something |= $built_sketch;
     $builder->run_hook('sketch.postbuild');
